/*
 ============================================================================
 Project     : Standard GA Algorithm
 Author      : Lucas Angelo Silveira
 Universidade de Brasilia (UnB) - Brazil
 Group of Theory of Computation
 Version     : 1.0
 
 Last Modified On: May 03, 2015.
 ============================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include "structures.h"
#include <math.h> //log2
#include "fitness/Processing.c"

float prob_crossover;     //Crossover Probability
float porc_selection;     //% of the Population that will be selected for Crossover
float porc_reemp;           //Pct. of the Population from where it will be performed
float prob_mutation;     //Mutation Probability
int * identity;
int size;
int sentinela;
int *c;
Chromossome *a;



void read_permutation_of_file(char* name, permutation *perm)
{
    int i,n;
    FILE *filePointer;
    if ((filePointer = fopen(name,"r")) == NULL){
        printf("Warning File not found\n");
    }
    
   else
    {
        //read size of permutation
        fscanf(filePointer,"%d",&n);
        perm->size2 = n;
        size=n;
        //alocate memory for perm->pi
        perm->pi = malloc(perm->size2*sizeof(int));
        //leer permutacion
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&perm->pi[i]);
        }
        fscanf(filePointer,"%d",&n);
        identity= (int *) calloc (n, sizeof (int));
        for(i=0; i<n; i++)
        {
            fscanf(filePointer,"%d",&identity[i]);
        }
        
	
    }//end  else
    fclose(filePointer);
}//end read permutation of archivo

void see_permutation(permutation *perm)
{
 int i;
 printf("Showing genome\n");
 for(i=0;i<perm->size2;i++)
	printf("%d,",perm->pi[i]);
 printf("\n");

}//end see_permutation


void see_Population(permutation *perm,Population *pob)
{
 int i,j;
  printf("displaying the generated population\n");
 for(i=0; i<pob->sizePopulation;i++ )
  {
	for(j=0;j<perm->size2;j++)
	{
            printf("%d,", pob->chromossomes[i].genes[j]);
		
	}
     printf("\n");
  }
}//end see_Population


// Procedure that generates the initial Population
 
void Population_init(Population *pob, permutation *perm){
	int i,j,r,n,prob;
	
    //Calc size of Population
	n = perm->size2;
    int adicional[n];
	pob->sizePopulation =  (int) (n * (log(n) / log(2))); //size of Population = (nlogn); log in base 2
	pob->basePopulation = 0;
	pob->topPopulation = pob->sizePopulation;
    
	pob->limiteSelection = pob->sizePopulation * porc_selection;//limit de Population what go to crossover
	pob->baseReplace = pob->sizePopulation * porc_reemp;//base de pob a the from of where will replace
    
	pob->baseDescendencia = pob->sizePopulation;
	pob->topeDescendencia = pob->sizePopulation + pob->limiteSelection;
    
	pob->totalsize = pob->topeDescendencia;
    
  
	pob->chromossomes = malloc(pob->totalsize*sizeof(Chromossome)); //alocar memoria para poblacion total
    if (pob->chromossomes == NULL)
        printf("pob->chromossomes NULL in Population_init(x,x) in Alg_Geneti_simple\n");
    
    for(i=0;i<pob->totalsize;i++)
    {
        pob->chromossomes[i].genes = malloc(n * sizeof(int));//alocar memoria para cada cromosoma
        if (pob->chromossomes[i].genes == NULL)
            printf("pob->chromosomes[i].genes NULL in Population_init(x,x) in Alg_Geneti_simple\n");
    }
    
	/*create Population a partir of  permutacion*/
       // A genom is see with a permutation of concatenation chroms 
        
	for(i=pob->basePopulation; i < pob->topPopulation; i++)
    {
            
		for(j = 0; j < perm->size2; j++)
        {
            if(perm->pi[j]!=0 && j!=0 && (j+1 < perm->size2) && perm->pi[j+1]!=0 && perm->pi[j-1]!=0  )
            {
                 r=rand(); 
		         prob = r % 2;//values between 0 and 1
		         if (prob == 1)//if prob 1, then genes is positive
			     {
                     pob->chromossomes[i].genes[j] = perm->pi[j];
                      adicional[j]=perm->pi[j];
			    
                 }
                 else//if prob 0, then genes is negative
                 {
			    
		             pob->chromossomes[i].genes[j] = perm->pi[j] * -1;
                     adicional[j]=perm->pi[j]*-1;
                 }
            }//end if
            else
            {
			pob->chromossomes[i].genes[j] = perm->pi[j];	
		    adicional[j]=perm->pi[j];
            }
               }//end for internal
           
		pob->chromossomes[i].Size = perm->size2;
	}//end for extern
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++){
		pob->chromossomes[i].Size = perm->size2;
	}


}//end create init Population


//calculate the tranlocation distance to each genome of the Population
void fitness_Population(Population *pob, int flag){
	int i;
    
	if (flag > 1){ //fitness over descendants
		for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
        {
             load_genoma( pob->chromossomes[i].genes, identity,pob->chromossomes[i].Size, size);
			 pob->chromossomes[i].fitness = distance(pob->chromossomes[i].genes,identity, pob->chromossomes[i].Size);
            
        }
	}
	else
         { // fitness over population
             for(i=pob->basePopulation; i < pob->sizePopulation; i++)
	     	{   
                    load_genoma( pob->chromossomes[i].genes, identity,pob->chromossomes[i].Size, size);
                    pob->chromossomes[i].fitness = distance(pob->chromossomes[i].genes,identity, pob->chromossomes[i].Size);
                
            }
        }
}


//flag(0) is initial Population or flag(1) if contains descendent Population
void see_fitness(Population *pob ,int flag)
{
    int i;
    if(!flag)
    {
        for(i=pob->basePopulation; i < pob->sizePopulation; i++)
            printf("distance is:%d\n",pob->chromossomes[i].fitness);
    }
    else
    {
    
        for(i=pob->baseDescendencia; i < pob->topeDescendencia; i++)
            printf("distance eh:%d\n",pob->chromossomes[i].fitness);
    }

}



//orders Population by fitness
void countingSort(Population *pob, int k){
	int i,j;
	
   
    if (!sentinela) {
    c = (int *) calloc (k, sizeof (int));
    }
	 
    if (!sentinela)
    {
        
        a = malloc(pob->sizePopulation*sizeof(Chromossome));
        if (a == NULL) printf("a NULL en countingSort() \n");
    
        for(i=0;i<pob->sizePopulation;i++)
        {
            a[i].genes = malloc(pob->chromossomes[0].Size * sizeof(int));//alocar memoria para cada cromosoma
            if (a[i].genes == NULL)
                printf("a[i].genes NULL en countingSort() \n");
        }
        
    }
    
    if (sentinela) {
        for (i=0; i<k; i++) 
            c[i]=0;
        
    }
    
    if (!sentinela) {
        sentinela=1;
    }
    
    /*make  a copy of Population*/
	for(i=0; i<pob->sizePopulation; i++){
		a[i].fitness = pob->chromossomes[i].fitness;
        for(j=0; j<pob->chromossomes[0].Size; j++)
            a[i].genes[j] = pob->chromossomes[i].genes[j];
	}
    
	/*counting sort*/
	for(i=0; i<k; i++){
		c[i] = 0;
	}
    
	for(i=0; i<pob->sizePopulation; i++){
		c[a[i].fitness] = c[a[i].fitness] + 1;
	}
    
	for(i=1; i<k; i++){
		c[i] = c[i] + c[i-1];
	}
    
	for(i=pob->sizePopulation-1; i>=0; i--){
        pob->chromossomes[c[a[i].fitness]-1].fitness = a[i].fitness;
        for(j=0; j<pob->chromossomes[0].Size; j++)
            pob->chromossomes[c[a[i].fitness]-1].genes[j] = a[i].genes[j];
        
        
		c[a[i].fitness] = c[a[i].fitness] - 1;
	}
    
    
   
}



//takes best solution
void selection(Population *pob,  int n_generetion, int * better_soluction){ 
	
    countingSort(pob,pob->chromossomes[0].Size+2);
    
    if (pob->chromossomes[0].fitness < *better_soluction)
        *better_soluction = pob->chromossomes[0].fitness;
    
}


void crossover(Population *pob, permutation *perm){
	int i,j,k,pos_pai1, pos_pai2, ponto_crossover;
	float prob;

	k = pob->baseDescendencia;
	
	for(i=pob->basePopulation; i<pob->limiteSelection; i+=2)
    {
		if (k+1 == pob->totalsize)
			break;
        
		
		pos_pai1 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
		pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
		
		for(j=0; j<10; j++){
			if (pos_pai1 != pos_pai2)
                   break;
			pos_pai2 = rand() % pob->limiteSelection;//valores between 0 e limiteSelection-1
		}
        
		prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
		if (prob <= prob_crossover)
        {//crossover
			ponto_crossover = rand() % perm->size2;//valores entre 0 e tamanho de permutation - 1
			//aplly single_point_crossover na primera part
			for(j=0; j<ponto_crossover; j++){
				pob->chromossomes[k].genes[j] = pob->chromossomes[pos_pai1].genes[j];
				pob->chromossomes[k+1].genes[j] = pob->chromossomes[pos_pai2].genes[j];
			}
			//aplly single_point_crossover int the second part
			for(j=ponto_crossover; j<perm->size2; j++){
				pob->chromossomes[k+1].genes[j] = pob->chromossomes[pos_pai1].genes[j];
				pob->chromossomes[k].genes[j] = pob->chromossomes[pos_pai2].genes[j];
			}
			k = k + 2;
		}//end if crossover
	}//end for iniit
	pob->topeDescendencia = k;//update tope descendence
}


void mutation(Population *pob, permutation *perm){
	int i,j;
	float prob;
    int cont_mutation=0;
	/*mutation about descendents, baseDescendence=0*/
	for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		for(j=0; j<perm->size2; j++)
        {
			prob = ((rand() % 100) + 1)/100.0; //de 0.01, 0.02 ... a 1.0
			
            if(prob <= prob_mutation)
            {//applied a gene mutation (flip sign)
				pob->chromossomes[i].genes[j] = pob->chromossomes[i].genes[j] * -1;
                cont_mutation++;
			}
            
		}
       
        	}//end for
  }



void substitution(Population *pob){
	int i,j,pos_reemp;
    int conta=0;
	/*descendants replace the current Population*/
    for(i=pob->baseDescendencia; i<pob->topeDescendencia; i++)
    {
		//seek position of substitution
		int tamanhoReemp = pob->topPopulation - pob->baseReplace;
		pos_reemp = (rand() % tamanhoReemp) + pob->baseReplace;//(values between 0 and tamanhoReemp-1)+baseReplace
		
        //apply substitution 
		if (pob->chromossomes[i].fitness < pob->chromossomes[pos_reemp].fitness)
        {
            conta++;
			//copy fitness
            pob->chromossomes[pos_reemp].fitness = pob->chromossomes[i].fitness;
            //copy genes
            for(j=0; j<pob->chromossomes[0].Size; j++)
                pob->chromossomes[pos_reemp].genes[j] = pob->chromossomes[i].genes[j];
            
        }
	}
    
}


//the parameters are not being used, but are mantanained for next modifications
void Genetic(float probcrossover, float probmutation, float probselection, float probreplace)
{
    
 int i;
 permutation perm;
sentinela=0;
 Population pob;
 int better_soluction;
 int num_generation;
 read_permutation_of_file("outgenome.txt", &perm);
 //verifica se is-cotail
    if (is_cotail(perm.pi,identity,perm.size2, size))
    {
       
        prob_crossover=0.9;
        prob_mutation=0.02;
        porc_selection=0.8;
        porc_reemp=1-0.7;
        
        Population_init(&pob, &perm);
        fitness_Population(&pob, 1);
        better_soluction = pob.chromossomes[0].fitness;
        num_generation = perm.size2;
        
        for(i=2; i<=num_generation; i++)
        {
            selection(&pob,i,&better_soluction);//seleccion:sorting population
            crossover(&pob, &perm);
            mutation(&pob, &perm);
            fitness_Population(&pob,i);
            substitution(&pob);
   
        
        }
     
        
         better_soluction = pob.chromossomes[0].fitness;
        printf("GA Solution:%d\n", better_soluction);
        
    }
    else
        printf("*****************Input is not COTAIL************************");
    
    
}//end function Genetic





